@@ -26,28 +26,33 @@
 #include <stdlib.h>
 #include <stdio.h>
 
-#include "chewing-utf8-util.h"
-#include "global.h"
-#include "bopomofo-private.h"
 #include "chewingutil.h"
-#include "userphrase-private.h"
 #include "choice-private.h"
-#include "dict-private.h"
-#include "tree-private.h"
-#include "pinyin-private.h"
 #include "private.h"
 #include "chewingio.h"
 #include "mod_aux.h"
 #include "global-private.h"
-#include "plat_path.h"
-#include "chewing-private.h"
-#include "key2pho-private.h"
 
+#ifdef WITH_RUST
+#   include "chewing_rs.h"
+#   include "chewing_internal.h"
+#else
+#include "global.h"
+#include "bopomofo-private.h"
+#include "chewing-utf8-util.h"
+#include "dict-private.h"
+#include "key2pho-private.h"
+#include "pinyin-private.h"
+#include "plat_path.h"
+#include "tree-private.h"
+#include "userphrase-private.h"
+#include "chewing-private.h"
 #if WITH_SQLITE3
 #    include "chewing-sql.h"
 #else
 #    include "hash-private.h"
 #endif
+#endif
 
 const char *const kb_type_str[] = {
     "KB_DEFAULT",
@@ -62,14 +67,24 @@ const char *const kb_type_str[] = {
     "KB_HANYU_PINYIN",
     "KB_THL_PINYIN",
     "KB_MPS2_PINYIN",
-    "KB_CARPALX"
+    "KB_CARPALX",
+    "KB_COLEMAK_DH_ANSI",
+    "KB_COLEMAK_DH_ORTH",
 };
 
+#ifdef WITH_RUST
+const char *const DICT_FILES[] = {
+    "tsi.dat",
+    "word.dat",
+    NULL,
+};
+#else
 const char *const DICT_FILES[] = {
     DICT_FILE,
     PHONE_TREE_FILE,
     NULL,
 };
+#endif
 
 const char *const SYMBOL_TABLE_FILES[] = {
     SYMBOL_TABLE_FILE,
@@ -193,7 +208,7 @@ CHEWING_API ChewingContext *chewing_new2(const char *syspath,
     chewing_Reset(ctx);
 
     if (syspath) {
-        strncpy(search_path, syspath, sizeof(search_path) - 1);
+        strncpy(search_path, syspath, ARRAY_SIZE(search_path) - 1);
     } else {
         ret = get_search_path(search_path, sizeof(search_path));
         if (ret) {
@@ -292,10 +307,15 @@ CHEWING_API ChewingContext *chewing_new()
 CHEWING_API int chewing_Reset(ChewingContext *ctx)
 {
     ChewingData *pgdata;
-    ChewingStaticData static_data;
+    ChewingStaticData staticData;
     ChewingConfigData old_config;
     void (*logger) (void *data, int level, const char *fmt, ...);
     void *loggerData;
+#ifdef WITH_RUST
+    const void *dict;
+    void *ce;
+    void *ue;
+#endif
 
     if (!ctx) {
         return -1;
@@ -306,17 +326,32 @@ CHEWING_API int chewing_Reset(ChewingContext *ctx)
 
     /* Backup old config and restore it after clearing pgdata structure. */
     old_config = pgdata->config;
-    static_data = pgdata->static_data;
+    staticData = pgdata->staticData;
     logger = pgdata->logger;
     loggerData = pgdata->loggerData;
+#ifdef WITH_RUST
+    dict = pgdata->dict;
+    ce = pgdata->ce;
+    ue = pgdata->ue;
+    if (pgdata->bopomofoData.editorWithKeymap != 0) {
+        FreePhoneticEditor(pgdata->bopomofoData.editorWithKeymap);
+    }
+#endif
     memset(pgdata, 0, sizeof(ChewingData));
     pgdata->config = old_config;
-    pgdata->static_data = static_data;
+    pgdata->staticData = staticData;
     pgdata->logger = logger;
     pgdata->loggerData = loggerData;
 
+#ifdef WITH_RUST
+    pgdata->dict = dict;
+    pgdata->ce = ce;
+    pgdata->ue = ue;
+    chewing_set_KBType(ctx, KB_DEFAULT);
+#else
     /* bopomofoData */
     memset(&(pgdata->bopomofoData), 0, sizeof(BopomofoData));
+#endif
 
     /* choiceInfo */
     memset(&(pgdata->choiceInfo), 0, sizeof(ChoiceInfo));
@@ -330,8 +365,8 @@ CHEWING_API int chewing_Reset(ChewingContext *ctx)
     pgdata->bFullShape = HALFSHAPE_MODE;
     pgdata->bSelect = 0;
     pgdata->nSelect = 0;
-    pgdata->PointStart = -1;
-    pgdata->PointEnd = 0;
+    pgdata->pointStart = -1;
+    pgdata->pointEnd = 0;
     pgdata->phrOut.nNumCut = 0;
     return 0;
 }
@@ -347,6 +382,18 @@ CHEWING_API int chewing_set_KBType(ChewingContext *ctx, int kbtype)
 
     LOG_API("kbtype = %d", kbtype);
 
+#ifdef WITH_RUST
+    if (pgdata->bopomofoData.editorWithKeymap != 0) {
+        FreePhoneticEditor(pgdata->bopomofoData.editorWithKeymap);
+    }
+    if (kbtype < KB_TYPE_NUM && kbtype >= 0) {
+        pgdata->bopomofoData.editorWithKeymap = NewPhoneticEditor(kbtype);
+        return 0;
+    } else {
+        pgdata->bopomofoData.editorWithKeymap = NewPhoneticEditor(KB_DEFAULT);
+        return -1;
+    }
+#else
     if (kbtype < KB_TYPE_NUM && kbtype >= 0) {
         ctx->data->bopomofoData.kbtype = kbtype;
         return 0;
@@ -354,6 +401,7 @@ CHEWING_API int chewing_set_KBType(ChewingContext *ctx, int kbtype)
         ctx->data->bopomofoData.kbtype = KB_DEFAULT;
         return -1;
     }
+#endif
 }
 
 CHEWING_API int chewing_get_KBType(const ChewingContext *ctx)
@@ -365,9 +413,11 @@ CHEWING_API int chewing_get_KBType(const ChewingContext *ctx)
     }
     pgdata = ctx->data;
 
-    LOG_API("kbtype = %d", ctx->data->bopomofoData.kbtype);
+    int kbtype = BopomofoKbType(&ctx->data->bopomofoData);
 
-    return ctx->data->bopomofoData.kbtype;
+    LOG_API("kbtype = %d", kbtype);
+
+    return kbtype;
 }
 
 CHEWING_API char *chewing_get_KBString(const ChewingContext *ctx)
@@ -379,18 +429,29 @@ CHEWING_API char *chewing_get_KBString(const ChewingContext *ctx)
     }
     pgdata = ctx->data;
 
-    LOG_API("KBString = %s", kb_type_str[ctx->data->bopomofoData.kbtype]);
+    int kbtype = BopomofoKbType(&ctx->data->bopomofoData);
 
-    return strdup(kb_type_str[ctx->data->bopomofoData.kbtype]);
+    LOG_API("KBString = %s", kb_type_str[kbtype]);
+
+    return strdup(kb_type_str[kbtype]);
 }
 
 CHEWING_API void chewing_delete(ChewingContext *ctx)
 {
     if (ctx) {
         if (ctx->data) {
+            if (ctx->data->bSelect) {
+                chewing_cand_close(ctx);
+            }
             TerminatePinyin(ctx->data);
             TerminateEasySymbolTable(ctx->data);
             TerminateSymbolTable(ctx->data);
+#ifdef WITH_RUST
+            FreePhoneticEditor(ctx->data->bopomofoData.editorWithKeymap);
+            UserGetPhraseEnd(ctx->data, NULL);
+            UserEnumeratePhraseEnd(ctx->data->phraseEnumIter);
+            ctx->data->phraseEnumIter = NULL;
+#endif
             TerminateUserphrase(ctx->data);
             TerminateTree(ctx->data);
             TerminateDict(ctx->data);
@@ -694,8 +755,7 @@ CHEWING_API void chewing_set_ChiEngMode(ChewingContext *ctx, int mode)
 
     if (mode == CHINESE_MODE || mode == SYMBOL_MODE) {
         // remove all data inside buffer as switching mode.
-        BopomofoRemoveAll(&(ctx->data->bopomofoData));
-        MakeOutputWithRtn(ctx->output, ctx->data, KEYSTROKE_ABSORB);
+        chewing_clean_bopomofo_buf(ctx);
         ctx->data->bChiSym = mode;
     }
 }
@@ -743,11 +803,40 @@ CHEWING_API int chewing_get_ShapeMode(const ChewingContext *ctx)
     return ctx->data->bFullShape;
 }
 
+CHEWING_API void chewing_set_autoLearn(ChewingContext *ctx, int mode)
+{
+    ChewingData *pgdata;
+
+    if (!ctx) {
+        return;
+    }
+    pgdata = ctx->data;
+
+    LOG_API("mode = %d", mode);
+
+    if (mode == AUTOLEARN_ENABLED || mode == AUTOLEARN_DISABLED)
+        ctx->data->bAutoLearn = mode;
+}
+
+CHEWING_API int chewing_get_autoLearn(const ChewingContext *ctx)
+{
+    const ChewingData *pgdata;
+
+    if (!ctx) {
+        return -1;
+    }
+    pgdata = ctx->data;
+
+    LOG_API("bAutoLearn = %d", ctx->data->bAutoLearn);
+
+    return ctx->data->bAutoLearn;
+}
+
 static void CheckAndResetRange(ChewingData *pgdata)
 {
-    if (pgdata->PointStart > -1) {
-        pgdata->PointStart = -1;
-        pgdata->PointEnd = 0;
+    if (pgdata->pointStart > -1) {
+        pgdata->pointStart = -1;
+        pgdata->pointEnd = 0;
     }
 }
 
@@ -806,13 +895,13 @@ CHEWING_API int chewing_handle_Space(ChewingContext *ctx)
      * - "space as selection" mode is disable
      * - mode is not CHINESE_MODE
      * - has incompleted bopomofo (space is needed to complete it)
+     * - pre-edit buffer is empty
      */
-    if (!pgdata->config.bSpaceAsSelection || pgdata->bChiSym != CHINESE_MODE || BopomofoIsEntering(&ctx->data->bopomofoData)) {
+    if (!pgdata->config.bSpaceAsSelection || pgdata->bChiSym != CHINESE_MODE ||
+        BopomofoIsEntering(&ctx->data->bopomofoData) || pgdata->chiSymbolBufLen == 0) {
         return chewing_handle_Default(ctx, ' ');
     }
 
-    CheckAndResetRange(pgdata);
-
     /*
      * space = right when the follogin conditions are true
      * 1. In select mode
@@ -822,10 +911,9 @@ CHEWING_API int chewing_handle_Space(ChewingContext *ctx)
      */
     if (pgdata->bSelect && ctx->output->pci->pageNo < ctx->output->pci->nPage - 1) {
         return chewing_handle_Right(ctx);
-    } else {
-        return chewing_handle_Down(ctx);
     }
-    return 0;
+
+    return chewing_handle_Down(ctx);
 }
 
 CHEWING_API int chewing_handle_Esc(ChewingContext *ctx)
@@ -880,34 +968,35 @@ CHEWING_API int chewing_handle_Enter(ChewingContext *ctx)
         keystrokeRtn = KEYSTROKE_IGNORE;
     } else if (pgdata->bSelect) {
         keystrokeRtn = KEYSTROKE_ABSORB | KEYSTROKE_BELL;
-    } else if (pgdata->PointStart > -1) {
+    } else if (pgdata->pointStart > -1) {
         int buf = pgdata->chiSymbolCursor;
-        int key;
+        int key = '0';
 
-        if (pgdata->PointEnd > 1) {
+        if (pgdata->pointEnd > 1) {
             if (!pgdata->config.bAddPhraseForward) {
-                pgdata->chiSymbolCursor = pgdata->PointStart;
-                key = '0' + pgdata->PointEnd;
+                pgdata->chiSymbolCursor = pgdata->pointStart;
             } else {
-                pgdata->chiSymbolCursor = pgdata->PointStart + pgdata->PointEnd;
-                key = '0' + pgdata->PointEnd;
+                pgdata->chiSymbolCursor = pgdata->pointStart + pgdata->pointEnd;
             }
-
-            chewing_handle_CtrlNum(ctx, key);
-            pgdata->chiSymbolCursor = buf;
-        } else if (pgdata->PointEnd < 1) {
+            key = '0' + pgdata->pointEnd;
+        } else if (pgdata->pointEnd < 1) {
             if (pgdata->config.bAddPhraseForward)
-                pgdata->chiSymbolCursor = buf - pgdata->PointEnd;
-            key = '0' - pgdata->PointEnd;
-            chewing_handle_CtrlNum(ctx, key);
-            pgdata->chiSymbolCursor = buf;
+                pgdata->chiSymbolCursor = buf - pgdata->pointEnd;
+            key = '0' - pgdata->pointEnd;
         }
-        pgdata->PointStart = -1;
-        pgdata->PointEnd = 0;
+        chewing_handle_CtrlNum(ctx, key);
+        pgdata->chiSymbolCursor = buf;
+        pgdata->pointStart = -1;
+        pgdata->pointEnd = 0;
+        MakeOutputWithRtn(pgo, pgdata, keystrokeRtn);
+        MakeOutputAddMsgAndCleanInterval(pgo, pgdata);
+        return 0;
     } else {
         keystrokeRtn = KEYSTROKE_COMMIT;
         WriteChiSymbolToCommitBuf(pgdata, pgo, nCommitStr);
-        AutoLearnPhrase(pgdata);
+        if (!pgdata->bAutoLearn) {
+            AutoLearnPhrase(pgdata);
+        }
         CleanAllBuf(pgdata);
         pgo->commitBufLen = nCommitStr;
     }
@@ -1071,16 +1160,16 @@ CHEWING_API int chewing_handle_ShiftLeft(ChewingContext *ctx)
         keystrokeRtn = KEYSTROKE_IGNORE;
     }
     if (!pgdata->bSelect) {
-        /*  PointEnd locates (-9, +9) */
-        if (!BopomofoIsEntering(&(pgdata->bopomofoData)) && pgdata->chiSymbolCursor > 0 && pgdata->PointEnd > -9) {
-            if (pgdata->PointStart == -1)
-                pgdata->PointStart = pgdata->chiSymbolCursor;
+        /*  pointEnd locates (-9, +9) */
+        if (!BopomofoIsEntering(&(pgdata->bopomofoData)) && pgdata->chiSymbolCursor > 0 && pgdata->pointEnd > -9) {
+            if (pgdata->pointStart == -1)
+                pgdata->pointStart = pgdata->chiSymbolCursor;
             pgdata->chiSymbolCursor--;
             if (ChewingIsChiAt(pgdata->chiSymbolCursor, pgdata)) {
-                pgdata->PointEnd--;
+                pgdata->pointEnd--;
             }
-            if (pgdata->PointEnd == 0)
-                pgdata->PointStart = -1;
+            if (pgdata->pointEnd == 0)
+                pgdata->pointStart = -1;
         }
     }
 
@@ -1142,17 +1231,17 @@ CHEWING_API int chewing_handle_ShiftRight(ChewingContext *ctx)
     }
 
     if (!pgdata->bSelect) {
-        /* PointEnd locates (-9, +9) */
+        /* pointEnd locates (-9, +9) */
         if (!BopomofoIsEntering(&(pgdata->bopomofoData)) &&
-            pgdata->chiSymbolCursor < pgdata->chiSymbolBufLen && pgdata->PointEnd < 9) {
-            if (pgdata->PointStart == -1)
-                pgdata->PointStart = pgdata->chiSymbolCursor;
+            pgdata->chiSymbolCursor < pgdata->chiSymbolBufLen && pgdata->pointEnd < 9) {
+            if (pgdata->pointStart == -1)
+                pgdata->pointStart = pgdata->chiSymbolCursor;
             if (ChewingIsChiAt(pgdata->chiSymbolCursor, pgdata)) {
-                pgdata->PointEnd++;
+                pgdata->pointEnd++;
             }
             pgdata->chiSymbolCursor++;
-            if (pgdata->PointEnd == 0)
-                pgdata->PointStart = -1;
+            if (pgdata->pointEnd == 0)
+                pgdata->pointStart = -1;
         }
     }
 
@@ -1221,7 +1310,8 @@ CHEWING_API int chewing_handle_Tab(ChewingContext *ctx)
         if (pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen) {
             pgdata->phrOut.nNumCut++;
             all_phrasing = 1;
-        } else if (ChewingIsChiAt(pgdata->chiSymbolCursor - 1, pgdata)) {
+        } else if (pgdata->chiSymbolCursor > 0 &&
+                   ChewingIsChiAt(pgdata->chiSymbolCursor - 1, pgdata)) {
             cursor = PhoneSeqCursor(pgdata);
             if (IsPreferIntervalConnted(cursor, pgdata)) {
                 pgdata->bUserArrBrkpt[cursor] = 1;
@@ -1408,7 +1498,7 @@ static int dvorak_convert(int key)
         'a', 'A', 'o', 'O', 'e', 'E', 'u', 'U', 'i', 'I', 'd', 'D', 'h', 'H',
         't', 'T', 'n', 'N', 's', 'S', '-', '_',
         ';', ':', 'q', 'Q', 'j', 'J', 'k', 'K', 'x', 'X', 'b', 'B', 'm', 'M',
-        'w', 'W', 'v', 'V', 'z', 'Z'
+        'w', 'W', 'v', 'V', 'z', 'Z', '[', '{', ']', '}'
     };
     const char qkey[] = {
         'q', 'Q', 'w', 'W', 'e', 'E', 'r', 'R', 't', 'T', 'y', 'Y', 'u', 'U',
@@ -1416,7 +1506,7 @@ static int dvorak_convert(int key)
         'a', 'A', 's', 'S', 'd', 'D', 'f', 'F', 'g', 'G', 'h', 'H', 'j', 'J',
         'k', 'K', 'l', 'L', ';', ':', '\'', '\"',
         'z', 'Z', 'x', 'X', 'c', 'C', 'v', 'V', 'b', 'B', 'n', 'N', 'm', 'M',
-        ',', '<', '.', '>', '/', '?'
+        ',', '<', '.', '>', '/', '?', '-', '_', '=', '+'
     };
     size_t i;
 
@@ -1467,7 +1557,7 @@ CHEWING_API int chewing_handle_Default(ChewingContext *ctx, int key)
     DEBUG_OUT("   key=%d", key);
 
     /* Dvorak Hsu */
-    if (pgdata->bopomofoData.kbtype == KB_DVORAK_HSU) {
+    if (BopomofoKbType(&pgdata->bopomofoData) == KB_DVORAK_HSU) {
         key = dvorak_convert(key);
     }
 
@@ -1536,7 +1626,8 @@ CHEWING_API int chewing_handle_Default(ChewingContext *ctx, int key)
                 keystrokeRtn = KEYSTROKE_ABSORB;
                 break;
             case BOPOMOFO_COMMIT:
-                AddChi(pgdata->bopomofoData.phone, pgdata->bopomofoData.phoneAlt, pgdata);
+                AddChi(BopomofoSyllableIndex(&pgdata->bopomofoData), BopomofoSyllableIndexAlt(&pgdata->bopomofoData), pgdata);
+                BopomofoRemoveAll(&pgdata->bopomofoData);
                 break;
             case BOPOMOFO_NO_WORD:
                 keystrokeRtn = KEYSTROKE_BELL | KEYSTROKE_ABSORB;
@@ -1633,12 +1724,6 @@ CHEWING_API int chewing_handle_CtrlNum(ChewingContext *ctx, int key)
     ChewingData *pgdata;
     ChewingOutput *pgo;
     int keystrokeRtn = KEYSTROKE_ABSORB;
-    int newPhraseLen;
-    int i;
-    uint16_t addPhoneSeq[MAX_PHONE_SEQ_LEN];
-    char addWordSeq[MAX_PHONE_SEQ_LEN * MAX_UTF8_SIZE + 1];
-    int phraseState;
-    int cursor;
 
     if (!ctx) {
         return -1;
@@ -1650,11 +1735,10 @@ CHEWING_API int chewing_handle_CtrlNum(ChewingContext *ctx, int key)
 
     CheckAndResetRange(pgdata);
 
-    if (pgdata->bSelect)
+    if (pgdata->bSelect || BopomofoIsEntering(&(pgdata->bopomofoData))) {
+        MakeOutputWithRtn(pgo, pgdata, keystrokeRtn);
         return 0;
-
-    CallPhrasing(pgdata, 0);
-    newPhraseLen = key - '0';
+    }
 
     if (key == '0' || key == '1') {
         pgdata->bSelect = 1;
@@ -1666,15 +1750,24 @@ CHEWING_API int chewing_handle_CtrlNum(ChewingContext *ctx, int key)
         return 0;
     }
 
-    cursor = PhoneSeqCursor(pgdata);
-    if (!pgdata->config.bAddPhraseForward) {
-        if (newPhraseLen >= 1 && cursor + newPhraseLen - 1 <= pgdata->nPhoneSeq) {
-            if (NoSymbolBetween(pgdata, cursor, cursor + newPhraseLen)) {
+    if (key >= '2' && key <= '9') {
+        int i;
+        int newPhraseLen = key - '0';
+        int phraseState = 0;
+        int cursor = PhoneSeqCursor(pgdata);
+        int key_buf_cursor = pgdata->chiSymbolCursor;
+        uint16_t addPhoneSeq[MAX_PHONE_SEQ_LEN];
+        char addWordSeq[MAX_PHONE_SEQ_LEN * MAX_UTF8_SIZE + 1];
+
+        if (!pgdata->config.bAddPhraseForward) {
+            if (cursor + newPhraseLen <= pgdata->nPhoneSeq &&
+                NoSymbolBetween(pgdata, key_buf_cursor, key_buf_cursor + newPhraseLen)) {
+
                 /* Manually add phrase to the user phrase database. */
                 memcpy(addPhoneSeq, &pgdata->phoneSeq[cursor], sizeof(uint16_t) * newPhraseLen);
                 addPhoneSeq[newPhraseLen] = 0;
 
-                copyStringFromPreeditBuf(pgdata, cursor, newPhraseLen, addWordSeq, sizeof(addWordSeq));
+                copyStringFromPreeditBuf(pgdata, key_buf_cursor, newPhraseLen, addWordSeq, sizeof(addWordSeq));
 
                 phraseState = UserUpdatePhrase(pgdata, addPhoneSeq, addWordSeq);
                 SetUpdatePhraseMsg(pgdata, addWordSeq, newPhraseLen, phraseState);
@@ -1683,15 +1776,15 @@ CHEWING_API int chewing_handle_CtrlNum(ChewingContext *ctx, int key)
                 for (i = 1; i < newPhraseLen; i++)
                     pgdata->bUserArrBrkpt[cursor + i] = 0;
             }
-        }
-    } else {
-        if (newPhraseLen >= 1 && cursor - newPhraseLen >= 0) {
-            if (NoSymbolBetween(pgdata, cursor - newPhraseLen, cursor)) {
+        } else {
+            if (cursor - newPhraseLen >= 0 &&
+                NoSymbolBetween(pgdata, key_buf_cursor - newPhraseLen, key_buf_cursor)) {
+
                 /* Manually add phrase to the user phrase database. */
                 memcpy(addPhoneSeq, &pgdata->phoneSeq[cursor - newPhraseLen], sizeof(uint16_t) * newPhraseLen);
                 addPhoneSeq[newPhraseLen] = 0;
 
-                copyStringFromPreeditBuf(pgdata, cursor - newPhraseLen, newPhraseLen, addWordSeq, sizeof(addWordSeq));
+                copyStringFromPreeditBuf(pgdata, key_buf_cursor - newPhraseLen, newPhraseLen, addWordSeq, sizeof(addWordSeq));
 
                 phraseState = UserUpdatePhrase(pgdata, addPhoneSeq, addWordSeq);
                 SetUpdatePhraseMsg(pgdata, addWordSeq, newPhraseLen, phraseState);
@@ -1701,11 +1794,22 @@ CHEWING_API int chewing_handle_CtrlNum(ChewingContext *ctx, int key)
                     pgdata->bUserArrBrkpt[cursor - newPhraseLen + i] = 0;
             }
         }
+
+        if (!phraseState) {
+            snprintf(pgdata->showMsg, sizeof(pgdata->showMsg),
+                "\xE5\x8A\xA0\xE8\xA9\x9E\xE5\xA4\xB1\xE6\x95\x97\xEF\xBC\x9A\xE5\xAD\x97\xE6\x95\xB8"
+                "\xE4\xB8\x8D\xE7\xAC\xA6\xE6\x88\x96\xE5\xA4\xBE\xE9\x9B\x9C\xE7\xAC\xA6\xE8\x99\x9F"
+                /* 加詞失敗：字數不符或夾雜符號 */);
+            pgdata->showMsgLen = 14;
+        }
+
+        CallPhrasing(pgdata, 0);
+        MakeOutputWithRtn(pgo, pgdata, keystrokeRtn);
+        MakeOutputAddMsgAndCleanInterval(pgo, pgdata);
+        return 0;
     }
-    CallPhrasing(pgdata, 0);
-    MakeOutputWithRtn(pgo, pgdata, keystrokeRtn);
-    MakeOutputAddMsgAndCleanInterval(pgo, pgdata);
-    return 0;
+
+    return -1;
 }
 
 CHEWING_API int chewing_handle_ShiftSpace(ChewingContext *ctx)
@@ -1846,27 +1950,32 @@ CHEWING_API int chewing_userphrase_enumerate(ChewingContext *ctx)
 {
     ChewingData *pgdata;
 
-#if WITH_SQLITE3
-    int ret;
-#endif
-
     if (!ctx) {
         return -1;
     }
 
+#ifdef WITH_RUST
+    UserEnumeratePhraseEnd(ctx->data->phraseEnumIter);
+    ctx->data->phraseEnumIter = UserEnumeratePhrase(ctx->data->ue);
+#else
+#if WITH_SQLITE3
+    int ret;
+#endif
+
     pgdata = ctx->data;
 
     LOG_API("");
 
 #if WITH_SQLITE3
-    assert(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT]);
-    ret = sqlite3_reset(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT]);
+    assert(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT]);
+    ret = sqlite3_reset(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT]);
     if (ret != SQLITE_OK) {
         LOG_ERROR("sqlite3_reset returns %d", ret);
         return -1;
     }
 #else
-    pgdata->static_data.userphrase_enum = FindNextHash(pgdata, NULL);
+    pgdata->staticData.userphrase_enum = FindNextHash(pgdata, NULL);
+#endif
 #endif
     return 0;
 }
@@ -1875,19 +1984,23 @@ CHEWING_API int chewing_userphrase_has_next(ChewingContext *ctx, unsigned int *p
 {
     ChewingData *pgdata;
 
+    if (!ctx || !phrase_len || !bopomofo_len) {
+        return 0;
+    }
+
+#ifdef WITH_RUST
+    return UserEnumerateHasNext(ctx->data->phraseEnumIter, phrase_len, bopomofo_len);
+#else
 #if WITH_SQLITE3
     int ret;
 #endif
 
-    if (!ctx || !phrase_len || !bopomofo_len) {
-        return 0;
-    }
     pgdata = ctx->data;
 
     LOG_API("");
 
 #if WITH_SQLITE3
-    ret = sqlite3_step(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT]);
+    ret = sqlite3_step(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT]);
     if (ret != SQLITE_ROW) {
         if (ret != SQLITE_DONE) {
             LOG_ERROR("sqlite3_step returns %d", ret);
@@ -1895,24 +2008,26 @@ CHEWING_API int chewing_userphrase_has_next(ChewingContext *ctx, unsigned int *p
         return 0;
     }
 
-    *phrase_len = strlen((const char *) sqlite3_column_text(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT],
+    *phrase_len = strlen((const char *) sqlite3_column_text(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT],
                                                             SQL_STMT_USERPHRASE[STMT_USERPHRASE_SELECT].column
                                                             [COLUMN_USERPHRASE_PHRASE])) + 1;
 
-    *bopomofo_len = GetBopomofoBufLen(sqlite3_column_int(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT],
+    *bopomofo_len = GetBopomofoBufLen(sqlite3_column_int(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT],
                                                          SQL_STMT_USERPHRASE[STMT_USERPHRASE_SELECT].column
                                                          [COLUMN_USERPHRASE_LENGTH]));
 
     return 1;
 #else
-    if (pgdata->static_data.userphrase_enum) {
-        *phrase_len = strlen(pgdata->static_data.userphrase_enum->data.wordSeq) + 1;
-        *bopomofo_len = BopomofoFromUintArray(NULL, 0, pgdata->static_data.userphrase_enum->data.phoneSeq);
+    if (pgdata->staticData.userphrase_enum) {
+        *phrase_len = strlen(pgdata->staticData.userphrase_enum->data.wordSeq) + 1;
+        *bopomofo_len = BopomofoFromUintArray(NULL, 0, pgdata->staticData.userphrase_enum->data.phoneSeq);
         return 1;
 
     }
     return 0;
 #endif
+#endif
+    return 0;
 }
 
 CHEWING_API int chewing_userphrase_get(ChewingContext *ctx,
@@ -1921,6 +2036,13 @@ CHEWING_API int chewing_userphrase_get(ChewingContext *ctx,
 {
     ChewingData *pgdata;
 
+    if (!ctx || !phrase_buf || !phrase_len || !bopomofo_buf || !bopomofo_len) {
+        return -1;
+    }
+
+#ifdef WITH_RUST
+    return UserEnumerateGet(ctx->data->phraseEnumIter, phrase_buf, phrase_len, bopomofo_buf, bopomofo_len);
+#else
 #if WITH_SQLITE3
     const char *phrase;
     int length;
@@ -1928,19 +2050,16 @@ CHEWING_API int chewing_userphrase_get(ChewingContext *ctx,
     uint16_t phone_array[MAX_PHRASE_LEN + 1] = { 0 };
 #endif
 
-    if (!ctx || !phrase_buf || !phrase_len || !bopomofo_buf || !bopomofo_len) {
-        return -1;
-    }
     pgdata = ctx->data;
 
     LOG_API("");
 
 #if WITH_SQLITE3
-    phrase = (const char *) sqlite3_column_text(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT],
+    phrase = (const char *) sqlite3_column_text(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT],
                                                 SQL_STMT_USERPHRASE[STMT_USERPHRASE_SELECT].column
                                                 [COLUMN_USERPHRASE_PHRASE]);
     length =
-        sqlite3_column_int(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT],
+        sqlite3_column_int(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT],
                            SQL_STMT_USERPHRASE[STMT_USERPHRASE_SELECT].column[COLUMN_USERPHRASE_LENGTH]);
 
     if (phrase_len < strlen(phrase) + 1) {
@@ -1954,7 +2073,7 @@ CHEWING_API int chewing_userphrase_get(ChewingContext *ctx,
     }
 
     for (i = 0; i < length && i < MAX_PHRASE_LEN; ++i) {
-        phone_array[i] = sqlite3_column_int(pgdata->static_data.stmt_userphrase[STMT_USERPHRASE_SELECT],
+        phone_array[i] = sqlite3_column_int(pgdata->staticData.stmt_userphrase[STMT_USERPHRASE_SELECT],
                                             SQL_STMT_USERPHRASE[STMT_USERPHRASE_SELECT].column[COLUMN_USERPHRASE_PHONE_0
                                                                                                + i]);
     }
@@ -1964,20 +2083,22 @@ CHEWING_API int chewing_userphrase_get(ChewingContext *ctx,
 
     return 0;
 #else
-    if (pgdata->static_data.userphrase_enum) {
-        strncpy(phrase_buf, pgdata->static_data.userphrase_enum->data.wordSeq, phrase_len);
+    if (pgdata->staticData.userphrase_enum) {
+        strncpy(phrase_buf, pgdata->staticData.userphrase_enum->data.wordSeq, phrase_len);
         phrase_buf[phrase_len - 1] = 0;
 
-        BopomofoFromUintArray(bopomofo_buf, bopomofo_len, pgdata->static_data.userphrase_enum->data.phoneSeq);
+        BopomofoFromUintArray(bopomofo_buf, bopomofo_len, pgdata->staticData.userphrase_enum->data.phoneSeq);
         bopomofo_buf[bopomofo_len - 1] = 0;
 
-        pgdata->static_data.userphrase_enum = FindNextHash(pgdata, pgdata->static_data.userphrase_enum);
+        pgdata->staticData.userphrase_enum = FindNextHash(pgdata, pgdata->staticData.userphrase_enum);
 
         return 0;
     }
 
     return -1;
 #endif
+#endif
+    return -1;
 }
 
 CHEWING_API int chewing_userphrase_add(ChewingContext *ctx, const char *phrase_buf, const char *bopomofo_buf)
@@ -2056,9 +2177,9 @@ CHEWING_API int chewing_userphrase_lookup(ChewingContext *ctx, const char *phras
     ssize_t phone_len;
     uint16_t *phone_buf = 0;
     int ret;
-    UserPhraseData *user_phrase_data;
+    UserPhraseData *user_phrase_data = NULL;
 
-    if (!ctx || !phrase_buf || !bopomofo_buf) {
+    if (!ctx || !bopomofo_buf) {
         return 0;
     }
     pgdata = ctx->data;
@@ -2077,7 +2198,7 @@ CHEWING_API int chewing_userphrase_lookup(ChewingContext *ctx, const char *phras
 
     user_phrase_data = UserGetPhraseFirst(pgdata, phone_buf);
     while (user_phrase_data) {
-        if (strcmp(phrase_buf, user_phrase_data->wordSeq) == 0)
+        if (phrase_buf == NULL || strcmp(phrase_buf, user_phrase_data->wordSeq) == 0)
             break;
         user_phrase_data = UserGetPhraseNext(pgdata, phone_buf);
     }
@@ -2302,7 +2423,9 @@ CHEWING_API int chewing_commit_preedit_buf(ChewingContext *ctx)
         return -1;
 
     WriteChiSymbolToCommitBuf(pgdata, pgo, len);
-    AutoLearnPhrase(pgdata);
+    if (!pgdata->bAutoLearn) {
+        AutoLearnPhrase(pgdata);
+    }
     CleanAllBuf(pgdata);
 
     MakeOutputWithRtn(pgo, pgdata, KEYSTROKE_COMMIT);
@@ -2352,3 +2475,13 @@ CHEWING_API int chewing_clean_bopomofo_buf(ChewingContext *ctx)
     MakeOutput(pgo, pgdata);
     return 0;
 }
+
+CHEWING_API int chewing_phone_to_bopomofo(unsigned short phone, char *buf, unsigned short len)
+{
+    if (!buf)
+        return GetPhoneLenFromUint(phone);
+    else if (len <= 0)
+        return -1;
+
+    return PhoneFromUint(buf, len, phone);
+}
